@@ -6,12 +6,12 @@
 
 This utility provides a conversion from WARC records to ZIM files.
 The WARCs are converted in a 'lossless' way, no data from WARC records is lost.
-Each WARC record results in two ZIM articles:
+Each WARC record results in two ZIM items:
 - The WARC payload is stored under /A/<url>
 - The WARC headers + HTTP headers are stored under the /H/<url>
 
 Given a WARC response record for 'https://example.com/',
-two ZIM articles are created /A/example.com/ and /H/example.com/ are created.
+two ZIM items are created /A/example.com/ and /H/example.com/ are created.
 
 Only WARC response and resource records are stored.
 
@@ -26,25 +26,32 @@ import json
 import pathlib
 import logging
 import tempfile
-import mimetypes
 import datetime
 import re
 import io
 import time
 from argparse import ArgumentParser
-from urllib.parse import urlsplit, urljoin, urlunsplit
+from urllib.parse import urlsplit, urljoin, urlunsplit, urldefrag, urlparse
 
 import pkg_resources
 import requests
+from libzim.writer import Hint
 from warcio import ArchiveIterator, StatusAndHeaders
 from warcio.recordbuilder import RecordBuilder
-from libzim.writer import Article, Blob
-from zimscraperlib.zim.creator import Creator
+from zimscraperlib.types import get_mime_for_name
 from zimscraperlib.i18n import setlocale, get_language_details, Locale
+from zimscraperlib.image.convertion import convert_image
+from zimscraperlib.image.transformation import resize_image
+from zimscraperlib.inputs import handle_user_provided_file
+from zimscraperlib.zim.creator import Creator
+from zimscraperlib.zim.items import StaticItem, URLItem
+from zimscraperlib.zim.providers import StringProvider
 from bs4 import BeautifulSoup
 
 from jinja2 import Environment, PackageLoader
 
+from cdxj_indexer import iter_file_or_dir, buffering_record_iter
+
 
 # Shared logger
 logger = logging.getLogger("warc2zim")
@@ -53,7 +60,7 @@ logger = logging.getLogger("warc2zim")
 HTML_TYPES = ("text/html", "application/xhtml", "application/xhtml+xml")
 
 
-# HTML raw mime type
+# HTML raw mime type (until https://github.com/kiwix/libkiwix/issues/671)
 HTML_RAW = "text/html;raw=true"
 
 
@@ -75,9 +82,10 @@ DEFAULT_TAGS = ["_ftindex:yes", "_category:other", "_sw:yes"]
 FUZZY_RULES = [
     {
         "match": re.compile(
-            r"//.*googlevideo.com/(videoplayback\?).*(id=[^&]+).*([&]itag=[^&]+).*"
+            # r"//.*googlevideo.com/(videoplayback\?).*(id=[^&]+).*([&]itag=[^&]+).*"
+            r"//.*googlevideo.com/(videoplayback\?).*((?<=[?&])id=[^&]+).*"
         ),
-        "replace": r"//youtube.fuzzy.replayweb.page/\1\2\3",
+        "replace": r"//youtube.fuzzy.replayweb.page/\1\2",
     },
     {
         "match": re.compile(
@@ -87,140 +95,108 @@ FUZZY_RULES = [
         "replace": r"//youtube.fuzzy.replayweb.page/\1\2",
     },
     {"match": re.compile(r"(\.[^?]+\?)[\d]+$"), "replace": r"\1"},
+    {
+        "match": re.compile(
+            r"//(?:www\.)?youtube(?:-nocookie)?\.com\/(youtubei\/[^?]+).*(videoId[^&]+).*"
+        ),
+        "replace": r"//youtube.fuzzy.replayweb.page/\1?\2",
+    },
+    {
+        "match": re.compile(r"//(?:www\.)?youtube(?:-nocookie)?\.com/embed/([^?]+).*"),
+        "replace": r"//youtube.fuzzy.replayweb.page/embed/\1",
+    },
+    {
+        "match": re.compile(
+            r".*(?:gcs-vimeo|vod|vod-progressive)\.akamaized\.net.*?/([\d/]+.mp4)$"
+        ),
+        "replace": r"vimeo-cdn.fuzzy.replayweb.page/\1",
+    },
+    {
+        "match": re.compile(r".*player.vimeo.com/(video/[\d]+)\?.*"),
+        "replace": r"vimeo.fuzzy.replayweb.page/\1",
+    },
 ]
 
 CUSTOM_CSS_URL = "https://warc2zim.kiwix.app/custom.css"
 
-
 # ============================================================================
-class BaseArticle(Article):
-    """BaseArticle for all ZIM Articles in warc2zim with default settings"""
-
-    def is_redirect(self):
-        return False
-
-    def get_title(self):
-        return ""
-
-    def get_filename(self):
-        return ""
-
-    def should_compress(self):
-        mime = self.get_mime_type()
-        return mime.startswith("text/") or mime in (
-            "application/warc-headers",
-            "application/javascript",
-            "application/json",
-            "image/svg+xml",
-        )
-
-    def should_index(self):
-        return False
-
-
-# ============================================================================
-class WARCHeadersArticle(BaseArticle):
-    """WARCHeadersArticle used to store the WARC + HTTP headers as text
+class WARCHeadersItem(StaticItem):
+    """WARCHeadersItem used to store the WARC + HTTP headers as text
     Usually stored under H namespace
     """
 
     def __init__(self, record):
         super().__init__()
         self.record = record
-        self.url = record.rec_headers.get("WARC-Target-URI")
+        self.url = get_record_url(record)
 
-    def get_url(self):
+    def get_path(self):
         return "H/" + canonicalize(self.url)
 
     def get_title(self):
         return ""
 
-    def get_mime_type(self):
+    def get_mimetype(self):
         return "application/warc-headers"
 
-    def get_data(self):
+    def get_hints(self):
+        return {Hint.FRONT_ARTICLE: False}
+
+    def get_contentprovider(self):
         # add WARC headers
         buff = self.record.rec_headers.to_bytes(encoding="utf-8")
         # add HTTP headers, if present
         if self.record.http_headers:
             buff += self.record.http_headers.to_bytes(encoding="utf-8")
 
-        return Blob(buff)
+        return StringProvider(content=buff, ref=self)
 
 
 # ============================================================================
-class WARCPayloadArticle(BaseArticle):
-    """WARCPayloadArticle used to store the WARC payload
+class WARCPayloadItem(StaticItem):
+    """WARCPayloadItem used to store the WARC payload
     Usually stored under A namespace
     """
 
     def __init__(self, record, head_insert=None, css_insert=None):
         super().__init__()
         self.record = record
-        self.url = record.rec_headers.get("WARC-Target-URI")
-        self.mime = get_record_mime_type(record)
+        self.url = get_record_url(record)
+        self.mimetype = get_record_mime_type(record)
         self.title = ""
-        self.payload = self.record.content_stream().read()
-        if self.mime == "text/html":
-            self.title = parse_title(self.payload)
-            if head_insert:
-                self.payload = HEAD_INS.sub(head_insert, self.payload)
-            if css_insert:
-                self.payload = CSS_INS.sub(css_insert, self.payload)
 
-    def get_url(self):
+        if hasattr(self.record, "buffered_stream"):
+            self.record.buffered_stream.seek(0)
+            self.content = self.record.buffered_stream.read()
+        else:
+            self.content = self.record.content_stream().read()
+
+        if self.mimetype == "text/html":
+            self.title = parse_title(self.content)
+            if head_insert:
+                self.content = HEAD_INS.sub(head_insert, self.content)
+            if css_insert:
+                self.content = CSS_INS.sub(css_insert, self.content)
+            self.mimetype = HTML_RAW
+
+    def get_path(self):
         return "A/" + canonicalize(self.url)
 
     def get_title(self):
         return self.title
 
-    def get_mime_type(self):
-        # converting text/html to application/octet-stream to avoid rewriting by kiwix
-        # original mime type still preserved in the headers block
-        return HTML_RAW if self.mime == "text/html" else self.mime
-
-    def get_data(self):
-        return Blob(self.payload)
-
-    def should_index(self):
-        return self.mime in HTML_TYPES
+    def get_hints(self):
+        return {}
 
 
 # ============================================================================
-class RemoteArticle(BaseArticle):
-    def __init__(self, filename, url):
-        super().__init__()
-        self.filename = filename
-        self.url = url
-
-        try:
-            resp = requests.get(url)
-            resp.raise_for_status()
-            self.content = resp.content
-            self.mime = resp.headers.get("Content-Type").split(";")[0]
-        except Exception as e:
-            logger.error(e)
-            logger.error("Unable to load URL: {0}".format(url))
-            raise
-
-    def get_url(self):
-        return "A/" + self.filename
-
-    def get_mime_type(self):
-        return self.mime
-
-    def get_data(self):
-        return Blob(self.content)
-
-
-# ============================================================================
-class StaticArticle(BaseArticle):
-    def __init__(self, env, filename, main_url):
-        super().__init__()
+class StaticArticle(StaticItem):
+    def __init__(self, env, filename, main_url, **kwargs):
+        super().__init__(**kwargs)
         self.filename = filename
         self.main_url = main_url
 
-        self.mime, _ = mimetypes.guess_type(filename)
+        self.mime = get_mime_for_name(filename)
         self.mime = self.mime or "application/octet-stream"
 
         if filename != SW_JS:
@@ -231,37 +207,14 @@ class StaticArticle(BaseArticle):
                 "warc2zim", "templates/" + filename
             ).decode("utf-8")
 
-    def get_url(self):
+    def get_path(self):
         return "A/" + self.filename
 
-    def get_mime_type(self):
+    def get_mimetype(self):
         return self.mime
 
-    def get_data(self):
-        return Blob(self.content.encode("utf-8"))
-
-
-# ============================================================================
-class RedirectArticle(BaseArticle):
-    def __init__(self, from_url, to_url):
-        super().__init__()
-        self.from_url = from_url
-        self.to_url = to_url
-
-    def get_url(self):
-        return self.from_url
-
-    def is_redirect(self):
-        return True
-
-    def get_redirect_url(self):
-        return self.to_url
-
-
-# ============================================================================
-class FaviconRedirectArticle(RedirectArticle):
-    def __init__(self, favicon_url):
-        super().__init__("-/favicon", "A/" + canonicalize(favicon_url))
+    def get_hints(self):
+        return {Hint.FRONT_ARTICLE: False}
 
 
 # ============================================================================
@@ -280,8 +233,9 @@ class WARC2Zim:
 
         if not self.zim_file:
             self.zim_file = "{name}_{period}.zim".format(
-                name=args.name, period=time.strftime("%Y-%m")
+                name=args.name, period="{period}"
             )
+        self.zim_file = self.zim_file.format(period=time.strftime("%Y-%m"))
 
         self.full_filename = os.path.join(self.output, self.zim_file)
 
@@ -289,7 +243,7 @@ class WARC2Zim:
         with tempfile.NamedTemporaryFile(dir=self.output, delete=True) as fh:
             logger.debug(f"Confirming output is writable using {fh.name}")
 
-        self.inputs = [pathlib.Path(path) for path in args.inputs]
+        self.inputs = args.inputs
         self.replay_viewer_source = args.replay_viewer_source
         self.custom_css = args.custom_css
 
@@ -321,7 +275,6 @@ class WARC2Zim:
             "scraper": "warc2zim " + get_version(),
         }
 
-        self.replay_articles = []
         self.revisits = {}
 
         # progress file handling
@@ -332,13 +285,29 @@ class WARC2Zim:
             self.stats_filename = self.output / self.stats_filename
         self.written_records = self.total_records = 0
 
-    def add_remote_or_local(self, filename):
-        if self.replay_viewer_source:
-            article = RemoteArticle(filename, self.replay_viewer_source + filename)
+    def add_replayer(self):
+        if self.replay_viewer_source and re.match(
+            r"^https?\:", self.replay_viewer_source
+        ):
+            self.creator.add_item(
+                URLItem(
+                    url=self.replay_viewer_source + SW_JS,
+                    path="A/" + SW_JS,
+                    mimetype="application/javascript",
+                )
+            )
+        elif self.replay_viewer_source:
+            self.creator.add_item_for(
+                fpath=self.replay_viewer_source + SW_JS,
+                path="A/" + SW_JS,
+                mimetype="application/javascript",
+            )
         else:
-            article = StaticArticle(self.env, filename, self.main_url)
-
-        self.replay_articles.append(article)
+            self.creator.add_item(
+                StaticArticle(
+                    self.env, SW_JS, self.main_url, mimetype="application/javascript"
+                )
+            )
 
     def init_env(self):
         # autoescape=False to allow injecting html entities from translated text
@@ -359,7 +328,7 @@ class WARC2Zim:
         return env
 
     def update_stats(self):
-        """ write progress as JSON to self.stats_filename if requested """
+        """write progress as JSON to self.stats_filename if requested"""
         if not self.stats_filename:
             return
         self.written_records += 1
@@ -422,63 +391,53 @@ class WARC2Zim:
         else:
             self.css_insert = None
 
-        self.add_remote_or_local(SW_JS)
-
-        for filename in pkg_resources.resource_listdir("warc2zim", "templates"):
-            if filename == HEAD_INSERT_FILE:
-                continue
-
-            if filename != SW_JS:
-                self.replay_articles.append(
-                    StaticArticle(self.env, filename, self.main_url)
-                )
-
-        self.total_records = sum(1 for _ in self.iter_warc_records())
-        logger.debug(f"Found {self.total_records} records in WARCs")
-
-        with Creator(
+        self.creator = Creator(
             self.full_filename,
-            main_page="index.html",
+            main_path="A/index.html",
             language=self.language or "eng",
             title=self.title,
             date=datetime.date.today(),
-            compression="zstd",
             **self.metadata,
-        ) as zimcreator:
-            # zimcreator.update_metadata(**self.metadata)
+        ).start()
 
-            for article in self.generate_all_articles():
-                if article:
-                    zimcreator.add_zim_article(article)
-                    if isinstance(article, WARCPayloadArticle):
-                        self.update_stats()
+        self.add_replayer()
 
-    def iter_warc_records(self, dir_iter=None):
+        for filename in pkg_resources.resource_listdir("warc2zim", "templates"):
+            if filename == HEAD_INSERT_FILE or filename == SW_JS:
+                continue
 
+            self.creator.add_item(StaticArticle(self.env, filename, self.main_url))
+
+        for record in self.iter_all_warc_records():
+            self.add_items_for_warc_record(record)
+
+        # process revisits, headers only
+        for url, record in self.revisits.items():
+            if canonicalize(url) not in self.indexed_urls:
+                logger.debug(
+                    "Adding revisit {0} -> {1}".format(
+                        url, record.rec_headers["WARC-Refers-To-Target-URI"]
+                    )
+                )
+                self.creator.add_item(WARCHeadersItem(record))
+                self.indexed_urls.add(url)
+
+        if self.favicon_url:
+            self.add_illustration()
+
+        logger.debug(f"Found {self.total_records} records in WARCs")
+
+        self.creator.finish()
+
+    def iter_all_warc_records(self):
+        # add custom css records
         if self.custom_css:
             yield self.get_custom_css_record()
 
-        curr_iter = dir_iter or iter(self.inputs)
-
-        for filename in curr_iter:
-            if filename.is_dir():
-                yield from self.iter_warc_records(filename.iterdir())
-                continue
-
-            # for directory iterator, only accept .warc, .warc.gz files
-            # (accept all files directly specified in self.inputs)
-            if dir_iter and not filename.name.endswith((".warc", ".warc.gz")):
-                continue
-
-            with open(filename, "rb") as warc_fh:
-                for record in ArchiveIterator(warc_fh):
-                    if record.rec_type not in ("resource", "response", "revisit"):
-                        continue
-
-                    yield record
+        yield from iter_warc_records(self.inputs)
 
     def find_main_page_metadata(self):
-        for record in self.iter_warc_records():
+        for record in self.iter_all_warc_records():
             if record.rec_type == "revisit":
                 continue
 
@@ -499,7 +458,7 @@ class WARC2Zim:
             ):
                 self.main_url = url
 
-            if self.main_url != url:
+            if urldefrag(self.main_url).url != url:
                 continue
 
             # if we get here, found record for the main page
@@ -513,7 +472,9 @@ class WARC2Zim:
                 )
                 return
 
-            content = record.content_stream().read()
+            # content = record.content_stream().read()
+            record.buffered_stream.seek(0)
+            content = record.buffered_stream.read()
 
             if not self.title:
                 self.title = parse_title(content)
@@ -566,36 +527,61 @@ class WARC2Zim:
                 self.language = lang_elem.attrs["content"]
                 return
 
-    def generate_all_articles(self):
-        # add replay system
-        for article in self.replay_articles:
-            yield article
-
-        for record in self.iter_warc_records():
-            yield from self.articles_for_warc_record(record)
-
-        # process revisits, headers only
-        for url, record in self.revisits.items():
-            if url not in self.indexed_urls:
-                logger.debug(
-                    "Adding revisit {0} -> {1}".format(
-                        url, record.rec_headers["WARC-Refers-To-Target-URI"]
-                    )
-                )
-                yield WARCHeadersArticle(record)
-                self.indexed_urls.add(url)
-
-        if not self.favicon_url:
+    def add_illustration(self, record=None):
+        if self.favicon_url in self.indexed_urls:
             return
 
-        if self.favicon_url not in self.indexed_urls:
-            logger.debug("Favicon not found in WARCs, fetching directly")
+        if record and record.http_headers.get_statuscode() != "200":
+            logger.warning(
+                "Favicon WARC record ({url}) is not usable. Skipping".format(
+                    url=record.rec_headers.get("WARC-Target-URI")
+                )
+            )
+            return
+
+        # add illustration from favicon option or in-warc favicon
+        if record is not None:
+            logger.info(
+                "Adding illustration from WARC record ({url})".format(
+                    url=record.rec_headers.get("WARC-Target-URI")
+                )
+            )
+            src_url = record.rec_headers.get("WARC-Target-URI")
+        else:
+            logger.info("Adding illustration from {url}".format(url=self.favicon_url))
+            src_url = self.favicon_url
+        favicon_fname = pathlib.Path(urlparse(src_url).path).name
+        src_illus_fpath = pathlib.Path(".").joinpath(favicon_fname)
+
+        # reusing payload from WARC record
+        if record:
+            with open(src_illus_fpath, "wb") as fh:
+                if hasattr(record, "buffered_stream"):
+                    record.buffered_stream.seek(0)
+                    fh.write(record.buffered_stream.read())
+                else:
+                    fh.write(record.content_stream().read())
+        # fetching online
+        else:
             try:
-                yield RemoteArticle(canonicalize(self.favicon_url), self.favicon_url)
-            except Exception:
+                handle_user_provided_file(source=self.favicon_url, dest=src_illus_fpath)
+            except Exception as exc:
+                logger.warning(
+                    "Unable to retrieve favicon. "
+                    "ZIM won't have an illustration: {exc}".format(exc=exc)
+                )
                 return
 
-        yield FaviconRedirectArticle(self.favicon_url)
+        # convert to PNG (might already be PNG but it's OK)
+        illus_fpath = src_illus_fpath.with_suffix(".png")
+        convert_image(src_illus_fpath, illus_fpath)
+
+        # resize to appropriate size (ZIM uses 48x48 so we double for retina)
+        for size in (96, 48):
+            resize_image(illus_fpath, width=size, height=size, method="thumbnail")
+            with open(illus_fpath, "rb") as fh:
+                self.creator.add_illustration(size, fh.read())
+        src_illus_fpath.unlink()
 
     def is_self_redirect(self, record, url):
         if record.rec_type != "response":
@@ -607,8 +593,12 @@ class WARC2Zim:
         location = record.http_headers["Location"]
         return canonicalize(url) == canonicalize(location)
 
-    def articles_for_warc_record(self, record):
-        url = record.rec_headers["WARC-Target-URI"]
+    def add_items_for_warc_record(self, record):
+        url = get_record_url(record)
+        if not url:
+            logger.debug(f"Skipping record with empty WARC-Target-URI {record}")
+            return
+
         if url in self.indexed_urls:
             logger.debug("Skipping duplicate {0}, already added to ZIM".format(url))
             return
@@ -627,13 +617,16 @@ class WARC2Zim:
                 logger.debug("Skipping self-redirect: " + url)
                 return
 
-            yield WARCHeadersArticle(record)
-            payload_article = WARCPayloadArticle(
-                record, self.head_insert, self.css_insert
-            )
+            self.creator.add_item(WARCHeadersItem(record))
+            payload_item = WARCPayloadItem(record, self.head_insert, self.css_insert)
 
-            if len(payload_article.payload) != 0:
-                yield payload_article
+            if len(payload_item.content) != 0:
+                self.creator.add_item(payload_item)
+                self.total_records += 1
+                self.update_stats()
+
+            if url == self.favicon_url:
+                self.add_illustration(record=record)
 
             self.indexed_urls.add(url)
 
@@ -667,6 +660,15 @@ class WARC2Zim:
         logger.debug("Adding fuzzy redirect {0} -> {1}".format(fuzzy_url, url))
 
 
+# ============================================================================
+def get_record_url(record):
+    """Check if record has url converted from POST/PUT, and if so, use that
+    otherwise return the target url"""
+    if hasattr(record, "urlkey"):
+        return record.urlkey
+    return record.rec_headers["WARC-Target-URI"]
+
+
 # ============================================================================
 def get_record_mime_type(record):
     if record.http_headers:
@@ -690,6 +692,16 @@ def parse_title(content):
         return ""
 
 
+# ============================================================================
+def iter_warc_records(inputs):
+    """iter warc records, including appending request data to matching response"""
+    for filename in iter_file_or_dir(inputs):
+        with open(filename, "rb") as fh:
+            for record in buffering_record_iter(ArchiveIterator(fh), post_append=True):
+                if record.rec_type in ("resource", "response", "revisit"):
+                    yield record
+
+
 # ============================================================================
 def warc2zim(args=None):
     parser = ArgumentParser(description="Create ZIM files from WARC files")
